# Copyright 2016 Sam Parkinson <sam@sam.today>
#
# This file is part of Something for Reddit.
#
# Something for Reddit is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Something for Reddit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Something for Reddit.  If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

from redditisgtk.markdownpango import markdown_to_pango, SaneLabel
from redditisgtk.palettebutton import connect_palette
from redditisgtk.api import get_reddit_api
from redditisgtk.buttons import (ScoreButtonBehaviour, AuthorButtonBehaviour,
                                 TimeButtonBehaviour, SubButtonBehaviour,
                                 process_shortcuts)


class CommentsView(Gtk.ScrolledWindow):
    '''Downloads comments, shows selftext'''

    got_post_data = GObject.Signal('got-post-data', arg_types=[object])

    def __init__(self, post=None, comments=None, permalink=None):
        Gtk.ScrolledWindow.__init__(self)
        self.add_events(Gdk.EventMask.KEY_PRESS_MASK)
        self.get_style_context().add_class('root-comments-view')
        self.props.hscrollbar_policy = Gtk.PolicyType.NEVER
        self._post = post
        self._msg = None

        self._permalink = permalink
        if post is not None and permalink is None:
            self._permalink = post['permalink']
        if self._permalink is None:
            raise Exception('We have no link for a post!')

        self._box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.add(self._box)
        self._box.show()

        self._spinner = Gtk.Spinner()
        self._box.pack_end(self._spinner, True, True, 0)
        self._spinner.show()

        if self._post is not None:
            self._init_post(self._post)

        if comments is not None:
            self.__message_done_cb(comments)
        else:
            self.refresh()

    def _init_post(self, post):
        self.got_post_data.emit(post)

        self._top = _PostTopBar(post, self, hideable=False,
                                refreshable=True, show_subreddit=True)
        self._top.get_style_context().add_class('root-comments-bar')
        self._box.add(self._top)
        self._top.show()

        selfpost_label = SaneLabel(
            '<big>{title}</big>\n'
            '{selftext_pango}'.format(
                selftext_pango=markdown_to_pango(post.get('selftext')),
                title=post['title']))
        selfpost_label.get_style_context().add_class('root-comments-label')
        self._box.add(selfpost_label)
        selfpost_label.show()

        self._comments = None
        self._load_full = None
        self._all_comments = [self._top]
        self._selected = self._top

    def focus(self):
        self._selected.grab_focus()

    def get_link_name(self):
        return self._post['name']

    def get_header_height(self):
        return self._comments.get_allocation().y

    def refresh(self, caller=None):
        self._spinner.start()
        self._msg = get_reddit_api().send_request(
            'GET', self._permalink, self.__message_done_cb)

    def do_unrealize(self):
        if self._msg is not None:
            get_reddit_api().cancel(self._msg)

    def __message_done_cb(self, j):
        self._spinner.stop()
        if self._post is None:
            self._post = j[0]['data']['children'][0]['data']
            self._init_post(self._post)

        self._msg = None
        self._all_comments = [self._top]
        self._selected = self._top

        if self._load_full is not None:
            self._load_full.hide()
            self._load_full.destroy()
        permalink_segments = len(self._permalink.rstrip('/').split('/'))
        # More than /r/rct/comments/4ns96b/new_to_rct_lovely_community_3/
        if permalink_segments > 6:
            self._load_full = LoadFullCommentsRow()
            self._load_full.load_full.connect(self.__load_full_cb)
            self._box.add(self._load_full)
            self._all_comments.append(self._load_full)
            self._load_full.show()
        elif self._load_full is not None:
            self._box.remove(self._load_full)
            self._load_full.destroy()

        if self._comments is not None:
            self._box.remove(self._comments)
            self._comments.hide()
            self._comments.destroy()

        if len(j[1]['data']['children']) == 0:
            self._comments = EmptyView('No Comments', action='Add a comment')
            self._comments.action.connect(self.__add_comment_clicked_cb)
        else:
            # The 0th one is just the self post
            self._comments = _CommentsView(j[1]['data']['children'],
                                           self,
                                           first=True)
        self._box.add(self._comments)
        self._comments.show()

    def __add_comment_clicked_cb(self, view):
        self._top.show_reply()

    def __load_full_cb(self, row):
        # First 6 = /r/rct/comments/4ns96b/new_to_rct_lovely_community_3/
        self._permalink = '/'.join(self._permalink.split('/')[:6])
        self.refresh()

    def reply_posted(self, new_id):
        # First 6 = /r/rct/comments/4ns96b/new_to_rct_lovely_community_3/
        parts = self._permalink.split('/')[:6]
        parts.append(new_id)
        self._permalink = '/'.join(parts)
        self.refresh()

    def get_original_poster(self):
        return self._post['author']

    def added_comment_row(self, row):
        self._all_comments.append(row)

    def do_event(self, event):
        def move(direction, jump):
            if jump:
                row = self._selected
                while isinstance(row, CommentRow) and row.depth > 0:
                    # FIXME:  This must be bad
                    # Me > List > Box > Revealer > Box > Next Row!!
                    row = row.get_parent().get_parent().get_parent() \
                             .get_parent().get_parent()
                kids = [self._top] + self._comments.get_children()
                i = kids.index(row)
                if 0 <= i + direction < len(kids):
                    row = kids[i + direction]
            else:
                i = self._all_comments.index(self._selected)
                row = None
                while 0 <= i + direction < len(self._all_comments):
                    row = self._all_comments[i + direction]
                    if row.get_mapped():
                        break
                    else:
                        i += direction
                        row = None

            if row is not None:
                row.grab_focus()
                self._selected = row
            else:
                # We went too far!
                self.error_bell()
                self._selected.get_style_context().remove_class('angry')
                self._selected.get_style_context().add_class('angry')
                GLib.timeout_add(
                    500,
                    self._selected.get_style_context().remove_class,
                    'angry')

        def load_full():
            if self._load_full is not None:
                self.__load_full_cb(None)

        shortcuts = {
            'k': (move, [-1, False]),
            'j': (move, [+1, False]),
            'Up': (move, [-1, False]),
            'Down': (move, [+1, False]),
            'h': (move, [-1, True]),
            'l': (move, [+1, True]),
            'Left': (move, [-1, True]),
            'Right': (move, [+1, True]),
            '<ctrl>f': (load_full, []),
            '<ctrl>r': (self.refresh, []),
        }
        return process_shortcuts(shortcuts, event)


class _CommentsView(Gtk.ListBox):
    '''
    Actually holds the comments

    This is recursive, eg, like:

    _CommentsView
        | CommentRow
        |   | _CommentsView
        |   |   | CommentRow
        | CommentRow
    '''

    def __init__(self, data, toplevel_cv, first=False, depth=0):
        Gtk.ListBox.__init__(self)
        self._first_row = None
        self._is_first = first
        self._depth = depth
        self._toplevel_cv = toplevel_cv

        ctx = self.get_style_context()
        ctx.add_class('comments-view')
        if self._is_first:
            ctx.add_class('first')
        ctx.add_class('depth-{}'.format(depth % 5))
        self._add_comments(data)

    def _add_comments(self, data):
        for comment in data:
            comment_data = comment['data']
            row = CommentRow(comment_data, self._depth, self._toplevel_cv)
            self._toplevel_cv.added_comment_row(row)
            row.recurse()
            row.got_more_comments.connect(self.__got_more_comments_cb)
            self.insert(row, -1)
            row.show()

    def do_row_activated(self, row):
        if self._is_first:
            viewport = self.get_parent().get_parent()
            comments_view = viewport.get_parent()
            # Stop the kinetic scroll, otherwise it will override our
            # scrolling adjustment
            comments_view.props.kinetic_scrolling = False

            # Scroll to the top of the collapsed row
            r = row.get_allocation()
            header = comments_view.get_header_height()
            adj = viewport.get_vadjustment()
            adj.props.value = r.y + header

            comments_view.props.kinetic_scrolling = True

            row.do_activated()

    def __got_more_comments_cb(self, caller_button, more_comments):
        caller_button.hide()
        self.remove(caller_button)
        caller_button.destroy()

        self._add_comments(more_comments)


class LoadFullCommentsRow(Gtk.ListBoxRow):

    load_full = GObject.Signal('load_full')

    def __init__(self):
        Gtk.ListBoxRow.__init__(self)

        self._ib = Gtk.InfoBar(message_type=Gtk.MessageType.QUESTION)
        self._ib.connect('response', self.__response_cb)
        self.add(self._ib)
        self._ib.show()

        label = Gtk.Label(label='Showing only a subset of comments')
        self._ib.get_content_area().add(label)
        label.show()

        self._button = self._ib.add_button('Show All Comments', 1)

    def grab_focus(self):
        self._button.grab_focus()

    def __response_cb(self, ib, response):
        self.load_full.emit()
        self._button.props.label = 'Loading...'


class _PostTopBar(Gtk.Bin):

    hide_toggled = GObject.Signal('hide-toggled', arg_types=[bool])

    def __init__(self, data, toplevel_cv, hideable=True,
                 refreshable=False, show_subreddit=False):
        Gtk.Bin.__init__(self, can_focus=True)
        self.add_events(Gdk.EventMask.KEY_PRESS_MASK)
        self.data = data
        self._toplevel_cv = toplevel_cv

        self._b = Gtk.Builder.new_from_resource(
            '/today/sam/reddit-is-gtk/post-top-bar.ui')
        self.add(self._b.get_object('box'))
        self.get_child().show()

        self.expand = self._b.get_object('expand')
        self.expand.props.visible = hideable
        self._b.get_object('refresh').props.visible = refreshable

        self._favorite = self._b.get_object('favorite')
        self._favorite.props.visible = 'saved' in self.data
        self._favorite.props.active = self.data.get('saved')

        self._name_button = self._b.get_object('name')
        self._abb = AuthorButtonBehaviour(
            self._name_button, self.data,
            self._toplevel_cv.get_original_poster(),
            show_flair=True)

        self._score_button = self._b.get_object('score')
        self._score_button.props.visible = 'score' in data
        if 'score' in data:
            self._sbb = ScoreButtonBehaviour(self._score_button, self.data)

        self._time_button = self._b.get_object('time')
        self._tbb = TimeButtonBehaviour(self._time_button, self.data)

        self._reply_button = self._b.get_object('reply')
        self._reply_pb = connect_palette(
            self._reply_button, self._make_reply_palette, recycle_palette=True)

        self._sub_button = self._b.get_object('sub')
        self._sub_button.props.visible = show_subreddit
        if show_subreddit:
            self._subbb = SubButtonBehaviour(self._sub_button, self.data)

        self._b.connect_signals(self)

        # We need to lazy allocate this list, otherwise we get bogus sizes
        self._hideables = None

    def do_get_request_mode(self):
        return Gtk.SizeRequestMode.HEIGHT_FOR_WIDTH

    def do_get_preferred_width(self):
        minimum, natural = Gtk.Bin.do_get_preferred_width(self)
        return 1, natural

    def do_get_preferred_height_for_width(self, width):
        # FIXME:  Worse for performance than the nested ListBoxes??
        if self._hideables is None:
            self._hideables = []
            for child in self._b.get_object('box').get_children():
                if child.props.visible:
                    minimum, natural = child.get_preferred_width()
                    self._hideables.append((child, natural))

        new_width = 0
        for b, w in self._hideables:
            new_width += w
            if new_width > width:
                b.hide()
            else:
                b.show()

        minh, nath = Gtk.Bin.do_get_preferred_height(self)
        return minh, nath

    def do_event(self, event):
        def toggle(button):
            button.props.active = not button.props.active

        def activate(button):
            button.props.active = True

        shortcuts = {
            'u': (self._sbb.vote, [+1]),
            'd': (self._sbb.vote, [-1]),
            'n': (self._sbb.vote, [0]),
            'f': (toggle, [self._favorite]),
            't': (activate, [self._time_button]),
            'a': (self.get_toplevel().goto_sublist,
                  ['/u/{}'.format(self.data['author'])]),
            's': (self.get_toplevel().goto_sublist,
                  ['/r/{}'.format(self.data['subreddit'])]),
            'r': (self.show_reply, []),
            'space': (toggle, [self.expand]),
            # The ListBoxRow usually eats these shortcuts, but we want
            # the ListBox to handle them, so we need to pass it up
            'Up': (self._toplevel_cv.do_event, [event]),
            'Down': (self._toplevel_cv.do_event, [event]),
        }
        return process_shortcuts(shortcuts, event)

    def show_reply(self):
        if self._reply_button.props.visible:
            self._reply_button.props.active = True
        else:
            self._show_reply_modal()

    def refresh_clicked_cb(self, button):
        self._toplevel_cv.refresh()

    def read_toggled_cb(self, toggle):
        if toggle.props.active:
            get_reddit_api().read_message(self.data['name'])
            self._read.get_style_context().remove_class('unread')
            self._read.props.active = True
            self._read.props.sensitive = False
            self._read.props.label = 'Read'

    def hide_toggled_cb(self, toggle):
        self.hide_toggled.emit(not toggle.props.active)

    def favorite_toggled_cb(self, button):
        get_reddit_api().set_saved(self.data['name'], button.props.active,
                                   None)

    def _make_reply_palette(self):
        popover = Gtk.Popover()
        contents = _ReplyPopoverContents(self.data)
        contents.posted.connect(self.__reply_posted_cb)
        popover.add(contents)
        return popover

    def _show_reply_modal(self):
        dialog = Gtk.Dialog(use_header_bar=True)
        contents = _ReplyPopoverContents(self.data,
                                         header_bar=dialog.get_header_bar())
        dialog.get_content_area().add(contents)
        contents.posted.connect(self.__reply_posted_cb)

        dialog.props.transient_for = self.get_toplevel()
        dialog.show()

    def __reply_posted_cb(self, caller, new_id):
        self._toplevel_cv.reply_posted(new_id)


class _ReplyPopoverContents(Gtk.Box):

    posted = GObject.Signal('posted', arg_types=[str])

    def __init__(self, data, header_bar=None, **kwargs):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)
        self.data = data

        sw = Gtk.ScrolledWindow()
        sw.set_size_request(500, 300)
        self.add(sw)
        self._textview = Gtk.TextView()
        self._textview.props.wrap_mode = Gtk.WrapMode.WORD
        self._textview.set_size_request(500, 300)
        self._textview.connect('event', self.__event_cb)
        sw.add(self._textview)

        self._done = Gtk.Button(label='Post Reply')
        self._done.connect('clicked', self.__done_clicked_cb)
        if header_bar is not None:
            header_bar.pack_end(self._done)
            self._done.get_style_context().add_class('suggested-action')
            self._done.show()
        else:
            self.add(self._done)

        self.show_all()

    def __event_cb(self, textview, event):
        shortcuts = {
            '<Ctrl>Return': (self.__done_clicked_cb, [None])
        }
        return process_shortcuts(shortcuts, event)

    def __done_clicked_cb(self, button):
        self._done.props.label = 'Sending...'
        self._done.props.sensitive = False
        b = self._textview.props.buffer
        text = b.get_text(b.get_start_iter(), b.get_end_iter(), False)
        get_reddit_api().reply(self.data['name'], text, self.__reply_done_cb)

    def __reply_done_cb(self, data):
        new_id = data['json']['data']['things'][0]['data']['id']
        self.posted.emit(new_id)

        parent = self.get_parent()
        parent.hide()
        parent.destroy()
        self.destroy()


class CommentRow(Gtk.ListBoxRow):

    refresh = GObject.Signal('refresh')
    got_more_comments = GObject.Signal('got-more-comments', arg_types=[object])

    def __init__(self, data, depth, toplevel_cv):
        Gtk.ListBoxRow.__init__(self, selectable=False)
        self.add_events(Gdk.EventMask.KEY_PRESS_MASK)
        self.data = data
        self.depth = depth
        self._toplevel_cv = toplevel_cv
        self._top = None
        self._more_button = None

        self._box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.add(self._box)
        self._box.show()

    def recurse(self):
        if 'body' in self.data:
            self._show_normal_comment()
        else:
            self._more_button = Gtk.Button.new_with_label(
                'Show {} more comments...'.format(self.data['count']))
            self._more_button.connect('clicked', self.__load_more_cb)
            self._more_button.get_style_context().add_class('load-more')
            self._box.add(self._more_button)
            self._more_button.show()

    def do_focus_in_event(self, event):
        if self._more_button is not None:
            self._more_button.grab_focus()

    def do_event(self, event):
        if self._top is not None:
            return self._top.do_event(event)

    def __load_more_cb(self, button):
        button.props.sensitive = False
        get_reddit_api().load_more(
            self._toplevel_cv.get_link_name(),
            self.data, self.__loaded_more_cb)

    def __loaded_more_cb(self, comments):
        self.got_more_comments.emit(comments)

    def _show_normal_comment(self):
        self._top = _PostTopBar(self.data, self._toplevel_cv)
        self._top.hide_toggled.connect(self.__hide_toggled_cb)
        self._box.add(self._top)
        self._top.show()

        body_pango = markdown_to_pango(self.data['body'])
        self._label = SaneLabel(body_pango)
        self._box.add(self._label)
        self._label.show()

        self._sub = None
        self._revealer = None
        if self.data.get('replies'):
            self._revealer = Gtk.Revealer(
                transition_type=Gtk.RevealerTransitionType.SLIDE_DOWN,
                reveal_child=True
            )
            revealer_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            self._revealer.add(revealer_box)
            revealer_box.show()
            self._box.add(self._revealer)
            self._revealer.show()

            self._sub = _CommentsView(
                self.data['replies']['data']['children'],
                self._toplevel_cv,
                depth=self.depth + 1)
            revealer_box.add(self._sub)
            self._sub.show()
        else:
            self._top.expand.hide()

    def __hide_toggled_cb(self, top, hidden):
        if self._revealer is not None:
            self._revealer.props.reveal_child = hidden

    def do_activated(self):
        if self._revealer is not None:
            rc = not self._revealer.props.reveal_child
            self._revealer.props.reveal_child = rc
            self._top.expand.props.active = not rc


class EmptyView(Gtk.Bin):

    action = GObject.Signal('action')

    def __init__(self, title, action=None):
        Gtk.Bin.__init__(self)

        builder = Gtk.Builder.new_from_resource(
            '/today/sam/reddit-is-gtk/empty-view.ui')
        self._g = builder.get_object
        self.add(self._g('box'))
        self.get_child().show()

        self._g('title').props.label = title

        if action is not None:
            self._g('action').props.label = action
            self._g('action').props.visible = True
            self._g('action').connect('clicked', self.__action_clicked_cb)

    def __action_clicked_cb(self, button):
        self.action.emit()
